# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from datetime import datetime
import json
import logging
import shutil
import tempfile

from mistral_lib import actions
from oslo_concurrency import processutils
from swiftclient import exceptions as swiftexceptions

from tripleo_common.actions import base
from tripleo_common import constants
from tripleo_common.utils import swift as swiftutils
from tripleo_common.utils import time_functions as timeutils

LOG = logging.getLogger(__name__)


class FormatMessagesAction(actions.Action):
    """Format messages as logs

    Given a list of Zaqar messages from the TripleO UI, return a log-formatted
    string
    """
    def __init__(self, messages):
        super(FormatMessagesAction, self).__init__()
        self.messages = messages

    def run(self, context):
        lines = []

        for zaqar_message in self.messages:
            log_object = zaqar_message.get('body')

            if not log_object:
                continue

            body = log_object.get('message', '')
            level = log_object.get('level', 'info')
            timestamp = log_object.get('timestamp', datetime.utcnow())

            if isinstance(body, (dict, list,)):
                body = json.dumps(body)

            lines.append(
                '{date} {level} {body}'.format(
                    date=timeutils.epoch_to_formatted_date(timestamp),
                    level=level,
                    body=body
                )
            )

        return '\n'.join(lines)


class PublishUILogToSwiftAction(base.TripleOAction):
    """Publish logs from UI to Swift"""

    def __init__(self, logging_data, logging_container):
        super(PublishUILogToSwiftAction, self).__init__()
        self.logging_data = logging_data
        self.logging_container = logging_container

    def _rotate(self, swift):
        """Optimistic log rotation

        Failure to sucessfully complete log rotation doesn't cause the
        entire action to fail
        """
        try:
            headers = swift.head_object(self.logging_container,
                                        constants.TRIPLEO_UI_LOG_FILENAME)
            if headers['content-length'] < constants.TRIPLEO_UI_LOG_FILE_SIZE:
                LOG.debug("Log file hasn't reached a full size so it doesn't"
                          " need to be rotated.")
                return
        except swiftexceptions.ClientException:
            LOG.debug("Couldn't get existing log file, skip log rotation.")
            return

        try:
            files = swift.get_container(self.logging_container)[1]
        except swiftexceptions.ClientException:
            LOG.warn("Logging container doesn't exist, skip log rotation.")
            return

        largest_existing_suffix = 0

        for f in files:
            try:
                suffix = int(f['name'].split('.')[-1])
                if suffix > largest_existing_suffix:
                    largest_existing_suffix = suffix
            except ValueError:
                continue

        next_suffix = largest_existing_suffix + 1
        next_filename = '{}.{}'.format(
            constants.TRIPLEO_UI_LOG_FILENAME, next_suffix)
        try:
            data = swift.get_object(self.logging_container,
                                    constants.TRIPLEO_UI_LOG_FILENAME)[1]
            swift.put_object(self.logging_container, next_filename, data)
            swift.delete_object(self.logging_container,
                                constants.TRIPLEO_UI_LOG_FILENAME)
        except swiftexceptions.ClientException as err:
            msg = "Log rotation failed: %s" % err
            LOG.warn(msg)

    def run(self, context):
        swift = self.get_object_client(context)
        swiftutils.get_or_create_container(swift, self.logging_container)
        self._rotate(swift)

        try:
            old_contents = swift.get_object(
                self.logging_container,
                constants.TRIPLEO_UI_LOG_FILENAME)[1]
            new_contents = old_contents + '\n' + self.logging_data
        except swiftexceptions.ClientException:
            LOG.debug(
                "There is no existing logging data, starting a new file.")
            new_contents = self.logging_data

        try:
            swift.put_object(self.logging_container,
                             constants.TRIPLEO_UI_LOG_FILENAME,
                             new_contents)
        except swiftexceptions.ClientException as err:
            msg = "Failed to publish logs: %s" % err
            return actions.Result(error=msg)


class PrepareLogDownloadAction(base.TripleOAction):
    """Publish all GUI logs to a temporary URL"""

    def __init__(self, logging_container, downloads_container, delete_after):
        super(PrepareLogDownloadAction, self).__init__()
        self.logging_container = logging_container
        self.downloads_container = downloads_container
        self.delete_after = delete_after

    def run(self, context):
        swift = self.get_object_client(context)
        tmp_dir = tempfile.mkdtemp()
        tarball_name = 'logs-%s.tar.gz' % timeutils.timestamp()

        try:
            swiftutils.download_container(
                swift, self.logging_container, tmp_dir)
            swiftutils.create_and_upload_tarball(
                swift, tmp_dir, self.downloads_container,
                tarball_name, self.delete_after)
        except swiftexceptions.ClientException as err:
            msg = "Error attempting an operation on container: %s" % err
            return actions.Result(error=msg)
        except (OSError, IOError) as err:
            msg = "Error while writing file: %s" % err
            return actions.Result(error=msg)
        except processutils.ProcessExecutionError as err:
            msg = "Error while creating a tarball: %s" % err
            return actions.Result(error=msg)
        except Exception as err:
            msg = "Error exporting logs: %s" % err
            return actions.Result(error=msg)
        finally:
            shutil.rmtree(tmp_dir)

        return tarball_name



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from mistral_lib import actions
from mistralclient.api import base as mistralclient_api
from oslo_concurrency.processutils import ProcessExecutionError

from tripleo_common.actions import base
from tripleo_common import constants
from tripleo_common.utils import nodes as nodeutils
from tripleo_common.utils import passwords as password_utils
from tripleo_common.utils import validations as utils


class GetSshKeyAction(base.TripleOAction):

    def run(self, context):
        mc = self.get_workflow_client(context)
        try:
            env = mc.environments.get('ssh_keys')
            p_key = env.variables[self.key_type]
        except Exception:
            ssh_key = password_utils.create_ssh_keypair()
            p_key = ssh_key[self.key_type]

            workflow_env = {
                'name': 'ssh_keys',
                'description': 'SSH keys for TripleO validations',
                'variables': ssh_key
            }
            mc.environments.create(**workflow_env)

        return p_key


class GetPubkeyAction(GetSshKeyAction):

    key_type = 'public_key'


class GetPrivkeyAction(GetSshKeyAction):

    key_type = 'private_key'


class Enabled(base.TripleOAction):
    """Indicate whether the validations have been enabled."""

    def _validations_enabled(self, context):
        """Detect whether the validations are enabled on the undercloud."""
        mistral = self.get_workflow_client(context)
        try:
            # NOTE: the `ssh_keys` environment is created by
            # instack-undercloud only when the validations are enabled on the
            # undercloud (or when they're installed manually). Therefore, we
            # can check for its presence here:
            mistral.environments.get('ssh_keys')
            return True
        except Exception:
            return False

    def run(self, context):
        return_value = {'stderr': ''}
        if self._validations_enabled(context):
            return_value['stdout'] = 'Validations are enabled'
            mistral_result = {"data": return_value}
        else:
            return_value['stdout'] = 'Validations are disabled'
            mistral_result = {"error": return_value}
        return actions.Result(**mistral_result)


class ListValidationsAction(base.TripleOAction):
    """Return a set of TripleO validations"""
    def __init__(self, groups=None):
        super(ListValidationsAction, self).__init__()
        self.groups = groups

    def run(self, context):
        return utils.load_validations(groups=self.groups)


class ListGroupsAction(base.TripleOAction):
    """Return a set of TripleO validation groups"""

    def run(self, context):
        validations = utils.load_validations()
        return {
            group for validation in validations
            for group in validation['groups']
        }


class RunValidationAction(base.TripleOAction):
    """Run the given validation"""
    def __init__(self, validation, plan=constants.DEFAULT_CONTAINER_NAME):
        super(RunValidationAction, self).__init__()
        self.validation = validation
        self.plan = plan

    def run(self, context):
        mc = self.get_workflow_client(context)
        identity_file = None
        try:
            env = mc.environments.get('ssh_keys')
            private_key = env.variables['private_key']
            identity_file = utils.write_identity_file(private_key)

            stdout, stderr = utils.run_validation(self.validation,
                                                  identity_file,
                                                  self.plan,
                                                  context)
            return_value = {'stdout': stdout, 'stderr': stderr}
            mistral_result = {"data": return_value}
        except mistralclient_api.APIException as e:
            return_value = {'stdout': '', 'stderr': e.error_message}
            mistral_result = {"error": return_value}
        except ProcessExecutionError as e:
            return_value = {'stdout': e.stdout, 'stderr': e.stderr}
            # Indicates to Mistral there was a failure
            mistral_result = {"error": return_value}
        finally:
            if identity_file:
                utils.cleanup_identity_file(identity_file)
        return actions.Result(**mistral_result)


class CheckBootImagesAction(base.TripleOAction):
    """Validate boot images"""

    # TODO(bcrochet): The validation actions are temporary. This logic should
    #                 move to the tripleo-validations project eventually.
    def __init__(self, images,
                 deploy_kernel_name=constants.DEFAULT_DEPLOY_KERNEL_NAME,
                 deploy_ramdisk_name=constants.DEFAULT_DEPLOY_RAMDISK_NAME):
        super(CheckBootImagesAction, self).__init__()
        self.images = images
        self.deploy_kernel_name = deploy_kernel_name
        self.deploy_ramdisk_name = deploy_ramdisk_name

    def run(self, context):
        messages = []
        kernel_id = self._check_for_image(self.deploy_kernel_name, messages)
        ramdisk_id = self._check_for_image(self.deploy_ramdisk_name, messages)

        return_value = {
            'kernel_id': kernel_id,
            'ramdisk_id': ramdisk_id,
            'errors': messages,
            'warnings': []
        }

        if messages:
            mistral_result = actions.Result(error=return_value)
        else:
            mistral_result = actions.Result(data=return_value)

        return mistral_result

    def _check_for_image(self, name, messages):
        multiple_message = ("Please make sure there is only one image named "
                            "'{}' in glance.")
        missing_message = ("No image with the name '{}' found - make sure you "
                           "have uploaded boot images.")

        image_id = None
        found_images = [item['id'] for item in self.images
                        if item['name'] == name]
        if len(found_images) > 1:
            messages.append(multiple_message.format(name))
        elif len(found_images) == 0:
            messages.append(missing_message.format(name))
        else:
            image_id = found_images[0]

        return image_id


class CheckFlavorsAction(base.TripleOAction):
    """Validate and collect nova flavors in use.

    Ensure that selected flavors (--ROLE-flavor) are valid in nova.
    Issue a warning if local boot is not set for a flavor.
    """

    # TODO(bcrochet): The validation actions are temporary. This logic should
    #                 move to the tripleo-validations project eventually.
    def __init__(self, roles_info):
        super(CheckFlavorsAction, self).__init__()
        self.roles_info = roles_info

    def run(self, context):
        """Validate and collect nova flavors in use.

        Ensure that selected flavors (--ROLE-flavor) are valid in nova.
        Issue a warning if local boot is not set for a flavor.

        :returns: dictionary flavor name -> (flavor object, scale)
        """
        compute_client = self.get_compute_client(context)
        flavors = {f.name: {'name': f.name, 'keys': f.get_keys()}
                   for f in compute_client.flavors.list()}

        result = {}
        warnings = []
        errors = []

        message = "Flavor '{1}' provided for the role '{0}', does not exist"
        warning_message = (
            'Flavor {0} "capabilities:boot_option" is set to '
            '"netboot". Nodes will PXE boot from the ironic '
            'conductor instead of using a local bootloader. '
            'Make sure that enough nodes are marked with the '
            '"boot_option" capability set to "netboot".')

        for target, (flavor_name, scale) in self.roles_info.items():
            if flavor_name is None or not scale:
                continue

            old_flavor_name, old_scale = result.get(flavor_name, (None, None))

            if old_flavor_name:
                result[flavor_name] = (old_flavor_name, old_scale + scale)
            else:
                flavor = flavors.get(flavor_name)

                if flavor:
                    keys = flavor.get('keys', None)
                    if keys:
                        if keys.get('capabilities:boot_option', '') \
                                == 'netboot':
                            warnings.append(
                                warning_message.format(flavor_name))

                    result[flavor_name] = (flavor, scale)
                else:
                    errors.append(message.format(target, flavor_name))

        return_value = {
            'flavors': result,
            'errors': errors,
            'warnings': warnings,
        }
        if errors:
            mistral_result = {'error': return_value}
        else:
            mistral_result = {'data': return_value}

        return actions.Result(**mistral_result)


class CheckNodeBootConfigurationAction(base.TripleOAction):
    """Check the boot configuration of the baremetal nodes"""

    # TODO(bcrochet): The validation actions are temporary. This logic should
    #                 move to the tripleo-validations project eventually.
    def __init__(self, node, kernel_id, ramdisk_id):
        super(CheckNodeBootConfigurationAction, self).__init__()

        self.node = node
        self.kernel_id = kernel_id
        self.ramdisk_id = ramdisk_id

    def run(self, context):
        warnings = []
        errors = []
        message = ("Node {uuid} has an incorrectly configured "
                   "{property}. Expected \"{expected}\" but got "
                   "\"{actual}\".")
        if self.node['driver_info'].get('deploy_ramdisk') != self.ramdisk_id:
            errors.append(message.format(
                uuid=self.node['uuid'],
                property='driver_info/deploy_ramdisk',
                expected=self.ramdisk_id,
                actual=self.node['driver_info'].get('deploy_ramdisk')
            ))
        if self.node['driver_info'].get('deploy_kernel') != self.kernel_id:
            errors.append(message.format(
                uuid=self.node['uuid'],
                property='driver_info/deploy_kernel',
                expected=self.kernel_id,
                actual=self.node['driver_info'].get('deploy_kernel')
            ))
        capabilities = nodeutils.capabilities_to_dict(
            self.node['properties'].get('capabilities', ''))
        if capabilities.get('boot_option') != 'local':
            boot_option_message = ("Node {uuid} is not configured to use "
                                   "boot_option:local in capabilities. It "
                                   "will not be used for deployment with "
                                   "flavors that require boot_option:local.")

            warnings.append(boot_option_message.format(uuid=self.node['uuid']))

        return_value = {
            'errors': errors,
            'warnings': warnings
        }
        if errors:
            mistral_result = {'error': return_value}
        else:
            mistral_result = {'data': return_value}

        return actions.Result(**mistral_result)


class VerifyProfilesAction(base.TripleOAction):
    """Verify that the profiles have enough nodes"""

    # TODO(bcrochet): The validation actions are temporary. This logic should
    #                 move to the tripleo-validations project eventually.
    def __init__(self, nodes, flavors):
        super(VerifyProfilesAction, self).__init__()

        self.nodes = nodes
        self.flavors = flavors

    def run(self, context):
        errors = []
        warnings = []

        bm_nodes = {node['uuid']: node for node in self.nodes
                    if node['provision_state'] in ('available', 'active')}

        if not bm_nodes:
            message = ('Error: There are no nodes in an available '
                       'or active state and with maintenance mode off.')
            return_value = {
                'errors': [message],
                'warnings': [],
            }
            return actions.Result(error=return_value)

        free_node_caps = {uu: self._node_get_capabilities(node)
                          for uu, node in bm_nodes.items()}

        profile_flavor_used = False
        for flavor_name, (flavor, scale) in self.flavors.items():
            if not scale:
                continue

            profile = None
            keys = flavor.get('keys')
            if keys:
                profile = keys.get('capabilities:profile')

            if not profile and len(self.flavors) > 1:
                message = ('Error: The {flavor} flavor has no profile '
                           'associated.\n'
                           'Recommendation: assign a profile with openstack '
                           'flavor set --property '
                           '"capabilities:profile"="PROFILE_NAME" {flavor}')

                errors.append(message.format(flavor=flavor_name))
                continue

            profile_flavor_used = True

            assigned_nodes = [uu for uu, caps in free_node_caps.items()
                              if caps.get('profile') == profile]
            required_count = scale - len(assigned_nodes)

            if required_count < 0:
                warnings.append('%d nodes with profile %s won\'t be used '
                                'for deployment now' % (-required_count,
                                                        profile))
                required_count = 0

            for uu in assigned_nodes:
                free_node_caps.pop(uu)

            if required_count > 0:
                message = ('Error: only {total} of {scale} requested ironic '
                           'nodes are tagged to profile {profile} (for flavor '
                           '{flavor}).\n'
                           'Recommendation: tag more nodes using openstack '
                           'baremetal node set --property "capabilities='
                           'profile:{profile},boot_option:local" <NODE ID>')
                errors.append(message.format(total=scale - required_count,
                                             scale=scale,
                                             profile=profile,
                                             flavor=flavor_name))

        nodes_without_profile = [uu for uu, caps in free_node_caps.items()
                                 if not caps.get('profile')]
        if nodes_without_profile and profile_flavor_used:
            warnings.append("There are %d ironic nodes with no profile that "
                            "will not be used: %s" % (
                                len(nodes_without_profile),
                                ', '.join(nodes_without_profile)))

        return_value = {
            'errors': errors,
            'warnings': warnings,
        }
        if errors:
            mistral_result = {'error': return_value}
        else:
            mistral_result = {'data': return_value}

        return actions.Result(**mistral_result)

    def _node_get_capabilities(self, node):
        """Get node capabilities."""
        return nodeutils.capabilities_to_dict(
            node['properties'].get('capabilities'))


class CheckNodesCountAction(base.TripleOAction):
    """Validate hypervisor statistics"""

    # TODO(bcrochet): The validation actions are temporary. This logic should
    #                 move to the tripleo-validations project eventually.
    def __init__(self, statistics, stack, associated_nodes, available_nodes,
                 parameters, default_role_counts):
        super(CheckNodesCountAction, self).__init__()
        self.statistics = statistics
        self.stack = stack
        self.associated_nodes = associated_nodes
        self.available_nodes = available_nodes
        self.parameters = parameters
        self.default_role_counts = default_role_counts

    def run(self, context):
        errors = []
        warnings = []

        requested_count = 0

        for param, default in self.default_role_counts.items():
            if self.stack:
                try:
                    current = int(self.stack['parameters'][param])
                except KeyError:
                    # We could be adding a new role on stack-update, so there's
                    # no assumption the parameter exists in the stack.
                    current = self.parameters.get(param, default)
                requested_count += self.parameters.get(param, current)
            else:
                requested_count += self.parameters.get(param, default)

        # We get number of nodes usable for the stack by getting already
        # used (associated) nodes and number of nodes which can be used
        # (not in maintenance mode).
        # Assumption is that associated nodes are part of the stack (only
        # one overcloud is supported).
        associated = len(self.associated_nodes)
        available = len(self.available_nodes)

        available_count = associated + available

        if requested_count > available_count:
            errors.append('Not enough baremetal nodes - available: %d, '
                          'requested: %d' %
                          (available_count, requested_count))

        if self.statistics['count'] < available_count:
            errors.append('Only %d nodes are exposed to Nova of %d requests. '
                          'Check that enough nodes are in "available" state '
                          'with maintenance mode off.' %
                          (self.statistics['count'], available_count))

        return_value = {
            'errors': errors,
            'warnings': warnings,
            'result': {
                'statistics': self.statistics,
                'enough_nodes': True,
                'requested_count': requested_count,
                'available_count': available_count,
            }
        }
        if errors:
            return_value['result']['enough_nodes'] = False
            mistral_result = {'error': return_value}
        else:
            mistral_result = {'data': return_value}

        return actions.Result(**mistral_result)



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import shutil
import tempfile
import yaml

from heatclient import exc as heatexceptions
from keystoneauth1 import exceptions as keystoneauth_exc
from mistral_lib import actions
from mistralclient.api import base as mistralclient_base
from oslo_concurrency import processutils
import six
from swiftclient import exceptions as swiftexceptions

from tripleo_common.actions import base
from tripleo_common import constants
from tripleo_common import exception
from tripleo_common.utils import plan as plan_utils
from tripleo_common.utils import swift as swiftutils
from tripleo_common.utils import tarball
from tripleo_common.utils.validations import pattern_validator


LOG = logging.getLogger(__name__)

default_container_headers = {
    constants.TRIPLEO_META_USAGE_KEY: 'plan'
}


class CreateContainerAction(base.TripleOAction):
    """Creates an object container

    This action creates an object container for a given name.  If a container
    with the same name already exists an exception is raised.
    """

    def __init__(self, container):
        super(CreateContainerAction, self).__init__()
        self.container = container

    def run(self, context):
        oc = self.get_object_client(context)

        # checks to see if a container has a valid name
        if not pattern_validator(constants.PLAN_NAME_PATTERN, self.container):
            message = ("Unable to create plan. The plan name must "
                       "only contain letters, numbers or dashes")
            return actions.Result(error=message)

        # checks to see if a container with that name exists
        if self.container in [container["name"] for container in
                              oc.get_account()[1]]:
            result_string = ("A container with the name %s already"
                             " exists.") % self.container
            return actions.Result(error=result_string)
        oc.put_container(self.container, headers=default_container_headers)


class MigratePlanAction(base.TripleOAction):
    """Migrate plan from using a Mistral environment to using Swift

    This action creates a plan-environment.yaml file based on the
    Mistral environment or the default environment, if the file doesn't
    already exist in Swift.

    This action will be deleted in Queens, as it will no longer be
    needed by then - all plans will include plan-environment.yaml by
    default.
    """

    def __init__(self, plan):
        super(MigratePlanAction, self).__init__()
        self.plan = plan

    def run(self, context):
        swift = self.get_object_client(context)
        mistral = self.get_workflow_client(context)
        from_mistral = False

        try:
            env = plan_utils.get_env(swift, self.plan)
        except swiftexceptions.ClientException:
            # The plan has not been migrated yet. Check if there is a
            # Mistral environment.
            try:
                env = mistral.environments.get(self.plan).variables
                from_mistral = True
            except (mistralclient_base.APIException,
                    keystoneauth_exc.http.NotFound):
                # No Mistral env and no template: likely deploying old
                # templates aka previous version of OpenStack.
                env = {'version': 1.0,
                       'name': self.plan,
                       'description': '',
                       'template': 'overcloud.yaml',
                       'environments': [
                           {'path': 'overcloud-resource-registry-puppet.yaml'}
                       ]}

            # Store the environment info into Swift
            plan_utils.put_env(swift, env)
            if from_mistral:
                mistral.environments.delete(self.plan)


class ListPlansAction(base.TripleOAction):
    """Lists deployment plans

    This action lists all deployment plans residing in the undercloud.  A
    deployment plan consists of a container marked with metadata
    'x-container-meta-usage-tripleo'.
    """

    def run(self, context):
        # Plans consist of a container object marked with metadata to ensure it
        # isn't confused with another container
        plan_list = []
        oc = self.get_object_client(context)

        for item in oc.get_account()[1]:
            container = oc.get_container(item['name'])[0]
            if constants.TRIPLEO_META_USAGE_KEY in container.keys():
                plan_list.append(item['name'])
        return list(set(plan_list))


class DeletePlanAction(base.TripleOAction):
    """Deletes a plan and associated files

    Deletes a plan by deleting the container matching plan_name. It
    will not delete the plan if a stack exists with the same name.

    Raises StackInUseError if a stack with the same name as plan_name
    exists.
    """

    def __init__(self, container):
        super(DeletePlanAction, self).__init__()
        self.container = container

    def run(self, context):
        error_text = None
        # heat throws HTTPNotFound if the stack is not found
        try:
            stack = self.get_orchestration_client(context).stacks.get(
                self.container
            )
        except heatexceptions.HTTPNotFound:
            pass
        else:
            if stack is not None:
                raise exception.StackInUseError(name=self.container)

        try:
            swift = self.get_object_client(context)
            swiftutils.delete_container(swift, self.container)
        except swiftexceptions.ClientException as ce:
            LOG.exception("Swift error deleting plan.")
            error_text = ce.msg
        except Exception as err:
            LOG.exception("Error deleting plan.")
            error_text = six.text_type(err)

        if error_text:
            return actions.Result(error=error_text)


class ListRolesAction(base.TripleOAction):
    """Returns a deployment plan's roles

    Parses roles_data.yaml and returns the names of all available roles.

    :param container: name of the Swift container / plan name
    :return: list of roles in the container's deployment plan
    """

    def __init__(self, container=constants.DEFAULT_CONTAINER_NAME):
        super(ListRolesAction, self).__init__()
        self.container = container

    def run(self, context):
        try:
            swift = self.get_object_client(context)
            roles_data = yaml.safe_load(swift.get_object(
                self.container, constants.OVERCLOUD_J2_ROLES_NAME)[1])
        except Exception as err:
            err_msg = ("Error retrieving roles data from deployment plan: %s"
                       % err)
            LOG.exception(err_msg)
            return actions.Result(error=err_msg)

        return [role['name'] for role in roles_data]


class ExportPlanAction(base.TripleOAction):
    """Exports a deployment plan

    This action exports a deployment plan with a given name. The plan
    templates are downloaded from the Swift container, packaged up in a tarball
    and uploaded to Swift.
    """

    def __init__(self, plan, delete_after, exports_container):
        super(ExportPlanAction, self).__init__()
        self.plan = plan
        self.delete_after = delete_after
        self.exports_container = exports_container

    def run(self, context):
        swift = self.get_object_client(context)
        tmp_dir = tempfile.mkdtemp()
        tarball_name = '%s.tar.gz' % self.plan

        try:
            swiftutils.download_container(swift, self.plan, tmp_dir)
            swiftutils.create_and_upload_tarball(
                swift, tmp_dir, self.exports_container, tarball_name,
                self.delete_after)
        except swiftexceptions.ClientException as err:
            msg = "Error attempting an operation on container: %s" % err
            return actions.Result(error=msg)
        except (OSError, IOError) as err:
            msg = "Error while writing file: %s" % err
            return actions.Result(error=msg)
        except processutils.ProcessExecutionError as err:
            msg = "Error while creating a tarball: %s" % err
            return actions.Result(error=msg)
        except Exception as err:
            msg = "Error exporting plan: %s" % err
            return actions.Result(error=msg)
        finally:
            shutil.rmtree(tmp_dir)


class UpdatePlanFromDirAction(base.TripleOAction):
    """Updates a plan and associated files

    Updates a plan by comparing the current files with the new ones
    provided:
        Updates only new files from the plan
        Add new files from the plan

    :param container: name of the Swift container / plan name
    """

    def __init__(self, container=constants.DEFAULT_CONTAINER_NAME,
                 templates_dir=constants.DEFAULT_TEMPLATES_PATH):
        super(UpdatePlanFromDirAction, self).__init__()
        self.container = container
        self.templates_dir = templates_dir

    def run(self, context):
        try:
            swift = self.get_object_client(context)
            # Upload template dir to tmp container
            container_tmp = '%s-tmp' % self.container
            with tempfile.NamedTemporaryFile() as tmp_tarball:
                tarball.create_tarball(self.templates_dir, tmp_tarball.name)
                tarball.tarball_extract_to_swift_container(
                    swift,
                    tmp_tarball.name,
                    container_tmp)
            # Get all new templates:
            new_templates = swift.get_object(container_tmp,
                                             '')[1].splitlines()
            old_templates = swift.get_object(self.container,
                                             '')[1].splitlines()
            exclude_user_data = [constants.PLAN_ENVIRONMENT,
                                 constants.OVERCLOUD_J2_ROLES_NAME,
                                 constants.OVERCLOUD_J2_NETWORKS_NAME,
                                 constants.OVERCLOUD_J2_EXCLUDES]
            # Update the old container
            for new in new_templates:
                # if doesn't exist, push it:
                if new not in old_templates:
                    swift.put_object(
                        self.container,
                        new,
                        swift.get_object(container_tmp, new)[1])
                else:
                    content_new = swift.get_object(container_tmp, new)
                    content_old = swift.get_object(self.container, new)
                    if (not content_new == content_old and
                       new not in exclude_user_data):
                        swift.put_object(
                            self.container,
                            new,
                            swift.get_object(container_tmp, new)[1])
        except swiftexceptions.ClientException as err:
            msg = "Error attempting an operation on container: %s" % err
            return actions.Result(error=msg)
        except Exception as err:
            msg = "Error while updating plan: %s" % err
            return actions.Result(error=msg)


class UpdatePlanEnvironmentAction(base.TripleOAction):
    """Updates the plan environment values

    Updates a plan environment values with the given parameters:
        Add new parameter
        Delete parameter

    :param parameter: key value of the parameter to add or delete
    :param value: value of the parameter to add or delete
    :param delete: True if the parameter should be deleted from the env
    :param env_key: environment key that should be one of the keys present
                    in the plan environment dictionary:
                         'passwords',
                         'description',
                         'parameter_defaults',
                         'environments',
                         'version',
                         'template',
                         'resource_registry',
                         'name'
    :param container: name of the Swift container / plan name
    """

    def __init__(self, parameter, env_key, value=None, delete=False,
                 container=constants.DEFAULT_CONTAINER_NAME):
        super(UpdatePlanEnvironmentAction, self).__init__()
        self.container = container
        self.parameter = parameter
        self.value = value
        self.delete = delete
        self.env_key = env_key

    def run(self, context):
        try:
            swift = self.get_object_client(context)
            plan_env = plan_utils.get_env(swift, self.container)
            if self.env_key in plan_env.keys():
                if self.delete:
                    try:
                        plan_env[self.env_key].pop(self.parameter)
                    except KeyError:
                        pass
                else:
                    plan_env[self.env_key].update({self.parameter: self.value})
            else:
                msg = "The environment key doesn't exist: %s" % self.env_key
                return actions.Result(error=msg)
        except swiftexceptions.ClientException as err:
            msg = "Error attempting an operation on container: %s" % err
            return actions.Result(error=msg)
        except Exception as err:
            msg = "Error while updating plan: %s" % err
            return actions.Result(error=msg)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml

from tripleo_common import constants


def update_in_env(swift, env, key, value='', delete_key=False):
    """Update plan environment."""
    if delete_key:
        try:
            del env[key]
        except KeyError:
            pass
    else:
        try:
            env[key].update(value)
        except (KeyError, AttributeError):
            env[key] = value

    put_env(swift, env)
    return env


def get_env(swift, name):
    """Get plan environment from Swift and convert it to a dictionary."""
    env = yaml.safe_load(
        swift.get_object(name, constants.PLAN_ENVIRONMENT)[1]
    )

    # Ensure the name is correct, as it will be used to update the
    # container later
    if env.get('name') != name:
        env['name'] = name

    return env


def put_env(swift, env):
    """Convert given environment to yaml and upload it to Swift."""
    swift.put_object(
        env['name'],
        constants.PLAN_ENVIRONMENT,
        yaml.safe_dump(env, default_flow_style=False)
    )



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
import mock
import yaml

from tripleo_common.constants import PLAN_NAME_PATTERN
from tripleo_common.tests import base
from tripleo_common.utils import validations


VALIDATION_GROUP_1 = """---
- hosts: overcloud
  vars:
    metadata:
      name: First validation
      description: A validation belonging to group1
      groups:
        - group1
  tasks:
  - name: Ping the nodes
    ping:
"""

VALIDATION_WITH_METADATA = """---
- hosts: undercloud
  vars:
    metadata:
      name: Validation with metadata
      description: A validation with extra metadata
      foo: a foo metadata
      bar: 42
  tasks:
  - name: Do something useful
    watch_tv:
"""

VALIDATION_GROUPS_1_2 = """---
- hosts: undercloud
  vars:
    metadata:
      name: Validation from many groups
      description: A validation belonging to groups 1 and 2
      groups:
        - group1
        - group2
  tasks:
  - name: Do something useful
    watch_tv:
"""

VALIDATION_GROUP_1_PARSED = {
    'description': 'A validation belonging to group1',
    'groups': ['group1'],
    'id': 'VALIDATION_GROUP_1',
    'metadata': {},
    'name': 'First validation',
}

VALIDATION_WITH_METADATA_PARSED = {
    'description': 'A validation with extra metadata',
    'groups': [],
    'id': 'VALIDATION_WITH_METADATA',
    'metadata': {'foo': 'a foo metadata', 'bar': 42},
    'name': 'Validation with metadata',
}

VALIDATION_GROUPS_1_2_PARSED = {
    'description': 'A validation belonging to groups 1 and 2',
    'groups': ['group1', 'group2'],
    'id': 'VALIDATION_GROUPS_1_2',
    'metadata': {},
    'name': 'Validation from many groups',
}


class ValidationsKeyTest(base.TestCase):

    @mock.patch("oslo_concurrency.processutils.execute")
    @mock.patch('tempfile.mkstemp')
    def test_write_identity_file(self, mock_mkstemp, mock_execute):
        mock_open_context = mock.mock_open()
        mock_mkstemp.return_value = 'fd', 'tmp_path'
        with mock.patch('os.fdopen',
                        mock_open_context):
            validations.write_identity_file('private_key')

        mock_open_context.assert_called_once_with('fd', 'w')
        mock_open_context().write.assert_called_once_with('private_key')
        mock_execute.assert_called_once_with(
            '/usr/bin/sudo', '/usr/bin/chown', 'validations:', 'tmp_path')

    @mock.patch("oslo_concurrency.processutils.execute")
    def test_cleanup_identity_file(self, mock_execute):
        validations.cleanup_identity_file('/path/to/key')
        mock_execute.assert_called_once_with(
            '/usr/bin/sudo', '/usr/bin/rm', '-f', '/path/to/key')


class LoadValidationsTest(base.TestCase):

    def test_get_validation_metadata(self):
        validation = yaml.safe_load(VALIDATION_GROUP_1)
        value = validations.get_validation_metadata(validation, 'name')
        self.assertEqual('First validation', value)

    @mock.patch('tripleo_common.utils.validations.DEFAULT_METADATA')
    def test_get_validation_metadata_default_value(self, mock_metadata):
        mock_metadata.get.return_value = 'default_value'
        value = validations.get_validation_metadata({}, 'missing')
        self.assertEqual('default_value', value)

    def test_get_remaining_metadata(self):
        validation = yaml.safe_load(VALIDATION_WITH_METADATA)
        value = validations.get_remaining_metadata(validation)
        expected = {
            'foo': 'a foo metadata',
            'bar': 42
        }
        self.assertEqual(expected, value)

    def test_get_remaining_metadata_no_extra(self):
        validation = yaml.safe_load(VALIDATION_GROUP_1)
        value = validations.get_remaining_metadata(validation)
        self.assertEqual({}, value)

    @mock.patch('glob.glob')
    def test_load_validations_no_group(self, mock_glob):
        mock_glob.return_value = ['VALIDATION_GROUP_1',
                                  'VALIDATION_WITH_METADATA']
        mock_open_context = mock.mock_open()
        mock_open_context().read.side_effect = [VALIDATION_GROUP_1,
                                                VALIDATION_WITH_METADATA]

        with mock.patch('tripleo_common.utils.validations.open',
                        mock_open_context):
            my_validations = validations.load_validations()

        expected = [VALIDATION_GROUP_1_PARSED, VALIDATION_WITH_METADATA_PARSED]
        self.assertEqual(expected, my_validations)

    @mock.patch('glob.glob')
    def test_load_validations_group(self, mock_glob):
        mock_glob.return_value = ['VALIDATION_GROUPS_1_2',
                                  'VALIDATION_GROUP_1',
                                  'VALIDATION_WITH_METADATA']
        mock_open_context = mock.mock_open()
        mock_open_context().read.side_effect = [VALIDATION_GROUPS_1_2,
                                                VALIDATION_GROUP_1,
                                                VALIDATION_WITH_METADATA]

        with mock.patch('tripleo_common.utils.validations.open',
                        mock_open_context):
            my_validations = validations.load_validations(groups=['group1'])

        expected = [VALIDATION_GROUPS_1_2_PARSED, VALIDATION_GROUP_1_PARSED]
        self.assertEqual(expected, my_validations)


class RunValidationTest(base.TestCase):

    @mock.patch('tripleo_common.utils.validations.find_validation')
    @mock.patch('oslo_concurrency.processutils.execute')
    def test_run_validation(self, mock_execute,
                            mock_find_validation):
        Ctx = namedtuple('Ctx', 'auth_uri user_name auth_token project_name')
        mock_ctx = Ctx(
            auth_uri='auth_uri',
            user_name='user_name',
            auth_token='auth_token',
            project_name='project_name'
        )
        mock_execute.return_value = 'output'
        mock_find_validation.return_value = 'validation_path'

        result = validations.run_validation('validation', 'identity_file',
                                            'plan', mock_ctx)
        self.assertEqual('output', result)
        mock_execute.assert_called_once_with(
            '/usr/bin/sudo', '-u', 'validations',
            'OS_AUTH_URL=auth_uri',
            'OS_USERNAME=user_name',
            'OS_AUTH_TOKEN=auth_token',
            'OS_TENANT_NAME=project_name',
            '/usr/bin/run-validation',
            'validation_path',
            'identity_file',
            'plan'
        )
        mock_find_validation.assert_called_once_with('validation')


class RunPatternValidatorTest(base.TestCase):

    def test_valid_patterns(self):
        self.assertTrue(validations.pattern_validator("^$", ""))
        self.assertTrue(
            validations.pattern_validator(PLAN_NAME_PATTERN, "foo"))
        self.assertTrue(
            validations.pattern_validator(PLAN_NAME_PATTERN, "Foo-1"))

    def test_invalid_patterns(self):
        self.assertFalse(
            validations.pattern_validator("^$", "foo"))
        self.assertFalse(
            validations.pattern_validator(PLAN_NAME_PATTERN, "foo_1"))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import fixtures
import mock
import os

from mock import call
from mock import patch

from tripleo_common.tests import base
from tripleo_common.tests.fake_config import fakes
from tripleo_common.utils import config as ooo_config


class TestConfig(base.TestCase):

    def setUp(self):
        super(TestConfig, self).setUp()

    @patch.object(ooo_config.Config, '_mkdir')
    @patch.object(ooo_config.Config, '_open_file')
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_overcloud_config_generate_config(self,
                                              mock_tmpdir,
                                              mock_open,
                                              mock_mkdir):
        config_type_list = ['config_settings', 'global_config_settings',
                            'logging_sources', 'monitoring_subscriptions',
                            'service_config_settings',
                            'service_metadata_settings',
                            'service_names',
                            'upgrade_batch_tasks', 'upgrade_tasks']
        fake_role = [role for role in
                     fakes.FAKE_STACK['outputs'][1]['output_value']]

        heat = mock.MagicMock()
        heat.stacks.get.return_value = fakes.create_tht_stack()
        self.config = ooo_config.Config(heat)
        mock_tmpdir.return_value = "/tmp/tht"
        self.config.download_config('overcloud', '/tmp', config_type_list)

        expected_mkdir_calls = [call('/tmp/tht/%s' % r) for r in fake_role]
        mock_mkdir.assert_has_calls(expected_mkdir_calls, any_order=True)
        expected_calls = []
        for config in config_type_list:
            for role in fake_role:
                if config == 'step_config':
                    expected_calls += [call('/tmp/tht/%s/%s.pp' %
                                            (role, config))]
                else:
                    expected_calls += [call('/tmp/tht/%s/%s.yaml' %
                                            (role, config))]
        mock_open.assert_has_calls(expected_calls, any_order=True)

    @patch.object(ooo_config.Config, '_mkdir')
    @patch.object(ooo_config.Config, '_open_file')
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_overcloud_config_one_config_type(self,
                                              mock_tmpdir,
                                              mock_open,
                                              mock_mkdir):

        expected_config_type = 'config_settings'
        fake_role = [role for role in
                     fakes.FAKE_STACK['outputs'][1]['output_value']]

        heat = mock.MagicMock()
        heat.stacks.get.return_value = fakes.create_tht_stack()
        self.config = ooo_config.Config(heat)
        mock_tmpdir.return_value = "/tmp/tht"
        self.config.download_config('overcloud', '/tmp', ['config_settings'])
        expected_mkdir_calls = [call('/tmp/tht/%s' % r) for r in fake_role]
        expected_calls = [call('/tmp/tht/%s/%s.yaml'
                          % (r, expected_config_type))
                          for r in fake_role]
        mock_mkdir.assert_has_calls(expected_mkdir_calls, any_order=True)
        mock_open.assert_has_calls(expected_calls, any_order=True)

    @mock.patch('os.mkdir')
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_overcloud_config_wrong_config_type(self, mock_tmpdir,
                                                mock_open, mock_mkdir):
        args = {'name': 'overcloud', 'config_dir': '/tmp',
                'config_type': ['bad_config']}
        heat = mock.MagicMock()
        heat.stacks.get.return_value = fakes.create_tht_stack()
        self.config = ooo_config.Config(heat)
        mock_tmpdir.return_value = "/tmp/tht"
        self.assertRaises(
            KeyError,
            self.config.download_config, *args)

    @mock.patch('tripleo_common.utils.config.Config.get_role_data',
                autospec=True)
    def test_overcloud_config_upgrade_tasks(self, mock_get_role_data):

        heat = mock.MagicMock()
        heat.stacks.get.return_value = fakes.create_tht_stack()
        self.config = ooo_config.Config(heat)
        self.tmp_dir = self.useFixture(fixtures.TempDir()).path
        fake_role = [role for role in
                     fakes.FAKE_STACK['outputs'][1]['output_value']]
        expected_tasks = {'FakeController': [{'name': 'Stop fake service',
                                              'service': 'name=fake '
                                              'state=stopped',
                                              'tags': 'step1',
                                              'when': 'step|int == 1'}],
                          'FakeCompute': [{'name': 'Stop fake service',
                                           'service':
                                           'name=fake state=stopped',
                                           'tags': 'step1',
                                           'when': ['step|int == 1',
                                                    'existingcondition']},
                                          {'name': 'Stop nova-'
                                           'compute service',
                                           'service':
                                           'name=openstack-nova-'
                                           'compute state=stopped',
                                           'tags': 'step1',
                                           'when': ['step|int == 1',
                                                    'existing', 'list']}]}
        mock_get_role_data.return_value = fake_role

        for role in fake_role:
            filedir = os.path.join(self.tmp_dir, role)
            os.makedirs(filedir)
            filepath = os.path.join(filedir, "upgrade_tasks_playbook.yaml")
            playbook_tasks = self.config._write_playbook_get_tasks(
                fakes.FAKE_STACK['outputs'][1]['output_value'][role]
                ['upgrade_tasks'], role, filepath)
            self.assertTrue(os.path.isfile(filepath))
            self.assertEqual(expected_tasks[role], playbook_tasks)



#   Copyright 2017 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import mock
import yaml

from tripleo_common.exception import NotFound
from tripleo_common.exception import RoleMetadataError
from tripleo_common.tests import base
from tripleo_common.utils import roles as rolesutils

SAMPLE_ROLE = """
###############################################################################
# Role: sample                                                                #
###############################################################################
- name: sample
  description: |
    Sample!
  networks:
    - InternalApi
  HostnameFormatDefault: '%stackname%-sample-%index%'
  ServicesDefault:
    - OS::TripleO::Services::Ntp
"""
SAMPLE_ROLE_OBJ = {
    'HostnameFormatDefault': '%stackname%-sample-%index%',
    'ServicesDefault': ['OS::TripleO::Services::Ntp'],
    'description': 'Sample!\n',
    'name': 'sample',
    'networks': ['InternalApi']
}


class TestRolesUtils(base.TestCase):
    @mock.patch('os.listdir')
    @mock.patch('os.path.exists')
    def test_get_roles_from_directory(self, exists_mock, listdir_mock):
        exists_mock.return_value = True
        listdir_mock.return_value = ['b.yaml', 'a.yaml']
        self.assertEqual(rolesutils.get_roles_list_from_directory('/foo'),
                         ['a', 'b'])

    @mock.patch('os.listdir')
    @mock.patch('os.path.exists')
    def test_get_roles_from_directory_failure(self, exists_mock, listdir_mock):
        exists_mock.return_value = False
        self.assertRaises(ValueError, rolesutils.get_roles_list_from_directory,
                          ['/foo'])

    def test_validate_roles(self):
        available_roles = ['a', 'b', 'c']
        requested_roles = ['b', 'c']
        try:
            rolesutils.check_role_exists(available_roles, requested_roles)
        except Exception:
            self.fail('Exception raised')

    def test_validate_roles_with_invalid_role(self):
        available_roles = ['a', 'b', 'c']
        requested_roles = ['b', 'd']
        self.assertRaises(NotFound, rolesutils.check_role_exists,
                          available_roles, requested_roles)

    @mock.patch('tripleo_common.utils.roles.check_role_exists')
    @mock.patch('tripleo_common.utils.roles.get_roles_list_from_directory')
    def test_generate_roles_data_from_directory(self, get_roles_mock,
                                                check_mock):
        get_roles_mock.return_value = ['foo', 'bar', 'baz']
        m = mock.mock_open(read_data=SAMPLE_ROLE)
        with mock.patch('tripleo_common.utils.roles.open', m) as open_mock:
            r = rolesutils.generate_roles_data_from_directory('/foo',
                                                              ['foo', 'bar'])
            open_mock.assert_any_call('/foo/foo.yaml', 'r')
            open_mock.assert_any_call('/foo/bar.yaml', 'r')

        header = '\n'.join(["#" * 79,
                            "# File generated by TripleO",
                            "#" * 79,
                            ""])
        expected = header + SAMPLE_ROLE * 2
        self.assertEqual(expected, r)
        get_roles_mock.assert_called_with('/foo')
        check_mock.assert_called_with(['foo', 'bar', 'baz'], ['foo', 'bar'])

    def test_validate_role_yaml(self):
        role = rolesutils.validate_role_yaml(SAMPLE_ROLE)
        self.assertEqual(SAMPLE_ROLE_OBJ, role)

    def test_validate_role_yaml_with_file(self):
        m = mock.mock_open(read_data=SAMPLE_ROLE)
        with mock.patch('tripleo_common.utils.roles.open', m):
            r = rolesutils.validate_role_yaml(role_path='/foo.yaml')
        self.assertEqual(SAMPLE_ROLE_OBJ, r)

    def test_validate_role_yaml_invalid_params(self):
        self.assertRaises(ValueError, rolesutils.validate_role_yaml, 'foo',
                          'bar')

    def test_validate_role_yaml_missing_name(self):
        role = yaml.safe_load(SAMPLE_ROLE)
        del role[0]['name']
        self.assertRaises(RoleMetadataError, rolesutils.validate_role_yaml,
                          yaml.dump(role))

    def test_validate_role_yaml_invalid_type(self):
        role = yaml.safe_load(SAMPLE_ROLE)
        role[0]['CountDefault'] = 'should not be a string'
        self.assertRaises(RoleMetadataError, rolesutils.validate_role_yaml,
                          yaml.dump(role))



# Copyright 2015 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fnmatch
import logging
import re
import time

import six

import heatclient.exc

LOG = logging.getLogger(__name__)


class DeployedServer(object):
    id = None
    name = None


class StackUpdateManager(object):
    def __init__(self, heatclient, novaclient, stack, hook_type,
                 nested_depth=5, hook_resource=None):
        self.heatclient = heatclient
        self.novaclient = novaclient
        self.stack = stack
        self.hook_type = hook_type
        self.nested_depth = nested_depth
        self.hook_resource = hook_resource
        self.server_names = {}
        self.servers = []

    def clear_breakpoints(self, refs):
        resources = self._resources_by_state()
        succeeds = []
        fails = []
        for ref in refs:
            server_name = None
            try:
                res = resources['on_breakpoint'][ref]
                server_name = self._server_name(ref)
                LOG.info("removing breakpoint on %s", server_name)
                stack_id = next(x['href'] for x in res.links if
                                x['rel'] == 'stack').rsplit('/', 1)[1]
                self.heatclient.resources.signal(
                    stack_id=stack_id,
                    resource_name=res.logical_resource_id,
                    data={'unset_hook': self.hook_type})
                succeeds.append(ref)
            except Exception as err:
                LOG.error("failed to remove breakpoint on %s: %s",
                          server_name or ref, err)
                fails.append(ref)
        return (succeeds, fails)

    def get_status(self):
        self.stack = self.heatclient.stacks.get(self.stack.id)
        # check if any of deployments' child resource has last
        # event indicating that it has reached a breakpoint (this
        # seems to be the only way how to check pre-create breakpoints ATM)
        resources = self._resources_by_state()
        if self.stack.status == 'IN_PROGRESS':
            if resources['on_breakpoint']:
                if resources['in_progress']:
                    status = 'IN_PROGRESS'
                else:
                    status = 'WAITING'
            else:
                status = 'IN_PROGRESS'
        else:
            status = self.stack.status
        LOG.debug('%s status: %s', self.stack.stack_name, status)
        return (status, resources)

    def do_interactive_update(self):
        status, _ = self.get_status()

        # wait for the stack-update to start
        while status in ['COMPLETE', 'FAILED']:
            status, _ = self.get_status()
            time.sleep(5)

        while status not in ['COMPLETE', 'FAILED']:
            status, resources = self.get_status()
            print(status)
            if status == 'WAITING':
                for state in resources:
                    if resources[state]:
                        print("{0}: {1}".format(state, self._server_names(
                                                resources[state].keys())))
                user_input = six.moves.input(
                    "Breakpoint reached, continue? Regexp or "
                    "Enter=proceed (will clear %s), "
                    "C-c=quit interactive mode: "
                    % resources['on_breakpoint'].keys()[-1])
                refs = self._input_to_refs(
                    user_input.strip(),
                    resources['on_breakpoint'].keys())
                self.clear_breakpoints(refs)
            time.sleep(5)
        print('update finished with status {0}'.format(status))

    def _resources_by_state(self):
        resources = {
            'not_started': {},
            'in_progress': {},
            'on_breakpoint': {},
            'completed': {},
            'failed': {},
        }
        all_resources = self.heatclient.resources.list(
            self.stack.id, nested_depth=self.nested_depth)
        if self.hook_type == 'pre-create':
            hook_reason = 'CREATE paused until Hook pre-create is cleared'
            hook_clear_reason = 'Hook pre-create is cleared'
        else:
            hook_reason = 'UPDATE paused until Hook pre-update is cleared'
            hook_clear_reason = 'Hook pre-update is cleared'

        stack_change_time = self._stack_change_time()

        for res in all_resources:
            if self.hook_resource:
                if not fnmatch.fnmatchcase(res.resource_name,
                                           self.hook_resource):
                    continue
            stack_name, stack_id = next(
                x['href'] for x in res.links if
                x['rel'] == 'stack').rsplit('/', 2)[1:]
            try:
                events = self.heatclient.events.list(
                    stack_id=stack_id,
                    resource_name=res.logical_resource_id,
                    sort_dir='asc')
            except heatclient.exc.HTTPNotFound:
                events = []
            state = 'not_started'
            for ev in events:
                # ignore events older than start of the last stack change
                if ev.event_time < stack_change_time:
                    continue
                if ev.resource_status_reason == hook_reason:
                    state = 'on_breakpoint'
                elif ev.resource_status_reason == hook_clear_reason:
                    state = 'in_progress'
                elif ev.resource_status in ('CREATE_IN_PROGRESS',
                                            'UPDATE_IN_PROGRESS'):
                    state = 'in_progress'
                elif ev.resource_status in ('CREATE_COMPLETE',
                                            'UPDATE_COMPLETE'):
                    state = 'completed'
            resources[state][res.physical_resource_id] = res

        return resources

    def _stack_change_time(self):
        if self.hook_type == 'pre-create':
            status_reason = 'Stack CREATE started'
        else:
            status_reason = 'Stack UPDATE started'
        events = self.heatclient.events.list(
            stack_id=self.stack.id,
            sort_dir='desc')
        try:
            ev = next(e for e in events if
                      e.resource_status_reason == status_reason)
            return ev.event_time
        except StopIteration:
            return None

    def _server_names(self, deployment_ids):
        return [self._server_name(i) for i in deployment_ids]

    def _server_name(self, deployment_id):
        name = self.server_names.get(deployment_id)
        if not name:
            if not self.servers:
                self.servers = self._get_servers()
            depl = self.heatclient.software_deployments.get(deployment_id)
            name = next(server.name for server in self.servers if
                        server.id == depl.server_id)
            self.server_names[deployment_id] = name
        return name

    def _get_servers(self):
        servers = self.novaclient.servers.list()

        # If no servers were found from Nova, we must be using split-stack,
        # so we will have to interrogate Heat for the names and id's.
        if not servers:
            resources = self.heatclient.resources.list(
                self.stack.id, nested_depth=self.nested_depth,
                filters=dict(type="OS::Heat::DeployedServer"))
            for res in resources:
                server = DeployedServer()
                stack_name, stack_id = next(
                    x['href'] for x in res.links if
                    x['rel'] == 'stack').rsplit('/', 2)[1:]
                stack = self.heatclient.stacks.get(stack_id)
                server.name = next(o['output_value'] for o in stack.outputs if
                                   o['output_key'] == 'name')
                server.id = res.physical_resource_id
                servers.append(server)

        return servers

    def _input_to_refs(self, regexp, refs):
        if regexp:
            try:
                pattern = "\A{0}\Z".format(regexp)
                return [ref for ref in refs if
                        re.match(pattern, self._server_name(ref))]
            except re.error as err:
                LOG.warning("'%s' is invalid regular expression: %s",
                            regexp.encode('string-escape'), err)
                return []
        else:
            return [refs.pop()]
